@@ -3,14 +3,17 @@
 
 from itertools import chain
 from typing import Optional, Any, Dict, Iterable, List, Union
+from copy import copy
 import warnings
 
-from .exceptions import *
+from .utils import short_name_as_id
+from .exceptions import DecodeError, OdxWarning
 from .globals import logger, xsi
 from .state import read_state_from_odx
 from .state_transition import read_state_transition_from_odx
 
-from .utils import make_ref, read_description_from_odx
+from .odxlink import OdxLinkRef, OdxLinkId,OdxLinkDatabase, OdxDocFragment
+from .utils import read_description_from_odx
 from .nameditemlist import NamedItemList
 from .admindata import AdminData, read_admin_data_from_odx
 from .companydata import CompanyData, read_company_datas_from_odx
@@ -39,15 +42,15 @@ class DiagLayer:
 
     class ParentRef:
         def __init__(self,
-                     reference,  # : Union[str, DiagLayer],
+                     parent : Union[OdxLinkRef, "DiagLayer"],
                      ref_type: str,
                      not_inherited_diag_comms=[],
                      not_inherited_dops=[]):
             """
             Parameters
             ----------
-            reference: str | DiagLayer
-                the ID (string) or the referenced DiagLayer
+            parent: OdxLinkRef | DiagLayer
+                A reference to the or the parent DiagLayer
             ref_type: str
             not_inherited_diag_comms: List[str]
                 short names of not inherited diag comms
@@ -57,41 +60,40 @@ class DiagLayer:
             if ref_type not in ["PROTOCOL-REF", "BASE-VARIANT-REF",
                                 "ECU-SHARED-DATA-REF", "FUNCTIONAL-GROUP-REF"]:
                 warnings.warn(f'Unknown parent ref type {ref_type}', OdxWarning)
-            if isinstance(reference, str):
-                self.id_ref = reference
-                self.referenced_diag_layer = None
+            if isinstance(parent, OdxLinkRef):
+                self.parent_ref = parent
+                self.parent_diag_layer = None
             else:
-                self.id_ref = reference.id
-                self.referenced_diag_layer = reference
+                assert isinstance(parent, DiagLayer)
+
+                self.parent_ref = OdxLinkRef.from_id(parent.odx_id)
+                self.parent_diag_layer = parent
             self.not_inherited_diag_comms = not_inherited_diag_comms
             self.not_inherited_dops = not_inherited_dops
             self.ref_type = ref_type
 
-        def _resolve_references(self, id_lookup):
-            self.referenced_diag_layer = id_lookup.get(self.id_ref)
-            if self.referenced_diag_layer is None:
-                logger.warning(
-                    f"Could not resolve parent ref to {self.id_ref}")
+        def _resolve_references(self, odxlinks: OdxLinkDatabase):
+            self.parent_diag_layer = odxlinks.resolve(self.parent_ref)
 
         def get_inheritance_priority(self):
-            return PRIORITY_OF_DIAG_LAYER_TYPE[self.referenced_diag_layer.variant_type]
+            return PRIORITY_OF_DIAG_LAYER_TYPE[self.parent_diag_layer.variant_type]
 
         def get_inherited_services_by_name(self):
-            services = {service.short_name: service for service in self.referenced_diag_layer._services
+            services = {service.short_name: service for service in self.parent_diag_layer._services
                         if service.short_name not in self.not_inherited_diag_comms}
             return services
 
         def get_inherited_data_object_properties_by_name(self):
-            dops = {dop.short_name: dop for dop in self.referenced_diag_layer._data_object_properties
+            dops = {dop.short_name: dop for dop in self.parent_diag_layer._data_object_properties
                     if dop.short_name not in self.not_inherited_dops}
             return dops
 
         def get_inherited_communication_parameters_by_name(self):
-            return {cp.id_ref: cp for cp in self.referenced_diag_layer._communication_parameters}
+            return {cp.id_ref.ref_id: cp for cp in self.parent_diag_layer._communication_parameters}
 
     def __init__(self,
                  variant_type,
-                 id,
+                 odx_id,
                  short_name,
                  long_name=None,
                  description=None,
@@ -100,13 +102,13 @@ class DiagLayer:
                  negative_responses: List[Response] = [],
                  services: List[DiagService] = [],
                  single_ecu_jobs: List[SingleEcuJob] = [],
-                 diag_comm_refs: List[str] = [],
+                 diag_comm_refs: List[OdxLinkRef] = [],
                  parent_refs: List[ParentRef] = [],
                  diag_data_dictionary_spec: DiagDataDictionarySpec = None,
                  communication_parameters:
                  List[CommunicationParameterRef] = [],
                  enable_candela_workarounds=True,
-                 id_lookup=None,
+                 odxlinks=None,
                  additional_audiences=[],
                  functional_classes=[],
                  states=[],
@@ -116,25 +118,25 @@ class DiagLayer:
         logger.info(f"Initializing variant type {variant_type}")
         self.variant_type = variant_type
 
-        self.id = id
+        self.odx_id = odx_id
         self.short_name = short_name
         self.long_name = long_name
         self.description = description
 
         # Requests and Responses
         self.requests = requests
-        self.positive_responses = NamedItemList[Response](lambda pr: pr.short_name,
+        self.positive_responses = NamedItemList[Response](short_name_as_id,
                                                           positive_responses)
-        self.negative_responses = NamedItemList[Response](lambda nr: nr.short_name,
+        self.negative_responses = NamedItemList[Response](short_name_as_id,
                                                           negative_responses)
 
         # ParentRefs
         self.parent_refs = parent_refs
 
         # DiagServices (note that they do not include inherited services!)
-        self._local_services = NamedItemList[DiagService](lambda ser: ser.short_name,
+        self._local_services = NamedItemList[DiagService](short_name_as_id,
                                                           services)
-        self._local_single_ecu_jobs = NamedItemList[SingleEcuJob](lambda job: job.short_name,
+        self._local_single_ecu_jobs = NamedItemList[SingleEcuJob](short_name_as_id,
                                                                   single_ecu_jobs)
         self._diag_comm_refs = diag_comm_refs
 
@@ -152,16 +154,16 @@ class DiagLayer:
 
         # Properties that include inherited objects
         self._services: NamedItemList[Union[DiagService, SingleEcuJob]]\
-            = NamedItemList(lambda s: s.short_name, [])
+            = NamedItemList(short_name_as_id, [])
         self._communication_parameters: NamedItemList[CommunicationParameterRef]\
-            = NamedItemList(lambda s: s.id_ref, [])
+            = NamedItemList(lambda cp: cp._python_name(), [])
         self._data_object_properties: NamedItemList[DopBase]\
-            = NamedItemList(lambda s: s.short_name, [])
+            = NamedItemList(short_name_as_id, [])
 
         self.import_refs = import_refs
 
-        if id_lookup is not None:
-            self.finalize_init(id_lookup)
+        if odxlinks is not None:
+            self.finalize_init(odxlinks)
 
         # specify whether enable work arounds for bugs of CANdela studio
         self._enable_candela_workarounds = enable_candela_workarounds
@@ -185,21 +187,26 @@ class DiagLayer:
         """All communication parameters including inherited ones."""
         return self._communication_parameters
 
-    def finalize_init(self, id_lookup={}):
+    def finalize_init(self, odxlinks: Optional[OdxLinkDatabase] = None):
         """Resolves all references.
 
         This method should be called whenever the diag layer (or a referenced object) was changed.
         Particularly, this method assumes that all inherited diag layer are correctly initialized,
         i.e., have resolved their references.
         """
-        id_lookup.update(self._build_id_lookup())
-        self._resolve_references(id_lookup)
 
-    def _build_id_lookup(self):
+        if odxlinks is None:
+            odxlinks = OdxLinkDatabase()
+
+        odxlinks.update(self._build_odxlinks())
+        self._resolve_references(odxlinks)
+
+    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
         """Construct a mapping from IDs to all objects that are contained in this diagnostic layer."""
-        logger.info(f"Adding {self.id} to id_lookup.")
+        logger.info(f"Adding {self.odx_id} to odxlinks.")
+
+        odxlinks = {}
 
-        id_lookup = {}
         for obj in chain(self._local_services,
                          self._local_single_ecu_jobs,
                          self.requests,
@@ -209,36 +216,36 @@ class DiagLayer:
                          self.functional_classes,
                          self.states,
                          self.state_transitions):
-            id_lookup[obj.id] = obj
+            odxlinks[obj.odx_id] = obj
 
         if self.local_diag_data_dictionary_spec:
-            id_lookup.update(
-                self.local_diag_data_dictionary_spec._build_id_lookup()
+            odxlinks.update(
+                self.local_diag_data_dictionary_spec._build_odxlinks()
             )
 
-        id_lookup[self.id] = self
-        return id_lookup
+        odxlinks[self.odx_id] = self
+        return odxlinks
 
-    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
+    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
         """Recursively resolve all references."""
         # Resolve inheritance
         for pr in self.parent_refs:
-            pr._resolve_references(id_lookup)
+            pr._resolve_references(odxlinks)
 
-        services = sorted(self._compute_available_services_by_name(id_lookup).values(),
-                          key=lambda service: service.short_name)
+        services = sorted(self._compute_available_services_by_name(odxlinks).values(),
+                          key=short_name_as_id)
         self._services = NamedItemList[Union[DiagService, SingleEcuJob]](
-            lambda s: s.short_name,
+            short_name_as_id,
             services)
 
         dops = sorted(self._compute_available_data_object_properties_by_name().values(),
-                      key=lambda dop: dop.short_name)
+                      key=short_name_as_id)
         self._data_object_properties = NamedItemList[DopBase](
-            lambda dop: dop.short_name,
+            short_name_as_id,
             dops)
 
         comparams = sorted(self._compute_available_commmunication_parameters_by_name().values(),
-                           key=lambda comparam: comparam.id_ref)
+                           key=lambda comparam: comparam.id_ref.ref_id)
         self._communication_parameters = NamedItemList[CommunicationParameterRef](
             lambda cp: cp._python_name(),
             comparams)
@@ -247,23 +254,23 @@ class DiagLayer:
         for struct in chain(self.requests,
                             self.positive_responses,
                             self.negative_responses):
-            struct._resolve_references(self, id_lookup)
+            struct._resolve_references(self, odxlinks)
 
         local_diag_comms: Iterable[Union[DiagService, SingleEcuJob]] \
             = (*self._local_services, *self._local_single_ecu_jobs)
         for service in local_diag_comms:
-            service._resolve_references(id_lookup)
+            service._resolve_references(odxlinks)
 
         if self.local_diag_data_dictionary_spec:
             self.local_diag_data_dictionary_spec._resolve_references(self,
-                                                                     id_lookup)
+                                                                     odxlinks)
 
-    def __local_services_by_name(self, id_lookup) -> Dict[str, Union[DiagService, SingleEcuJob]]:
+    def __local_services_by_name(self, odxlinks: OdxLinkDatabase) -> Dict[str, Union[DiagService, SingleEcuJob]]:
         services_by_name: Dict[str, Union[DiagService, SingleEcuJob]] = {}
 
         for ref in self._diag_comm_refs:
-            if ref in id_lookup:
-                services_by_name[id_lookup[ref].short_name] = id_lookup[ref]
+            if (obj := odxlinks.resolve_lenient(ref)) is not None:
+                services_by_name[obj.short_name] = obj
             else:
                 logger.warning(f"Diag comm ref {ref!r} could not be resolved.")
 
@@ -275,7 +282,7 @@ class DiagLayer:
         })
         return services_by_name
 
-    def _compute_available_services_by_name(self, id_lookup) -> Dict[str, DiagService]:
+    def _compute_available_services_by_name(self, odxlinks: OdxLinkDatabase) -> Dict[str, DiagService]:
         """Helper method for initializing the available services.
         This computes the services that are inherited from other diagnostic layers."""
         services_by_name = {}
@@ -286,7 +293,7 @@ class DiagLayer:
             services_by_name.update(
                 parent_ref.get_inherited_services_by_name())
 
-        services_by_name.update(self.__local_services_by_name(id_lookup))
+        services_by_name.update(self.__local_services_by_name(odxlinks))
         return services_by_name
 
     def _compute_available_data_object_properties_by_name(self) -> Dict[str, DopBase]:
@@ -315,11 +322,11 @@ class DiagLayer:
                 parent_ref.get_inherited_communication_parameters_by_name())
 
         com_params_by_name.update(
-            {cp.id_ref: cp for cp in self._local_communication_parameters})
+            {cp.id_ref.ref_id: cp for cp in self._local_communication_parameters})
         return com_params_by_name
 
     def _get_parent_refs_sorted_by_priority(self, reverse=False):
-        return sorted(self.parent_refs, key=lambda pr:  pr.get_inheritance_priority(), reverse=reverse)
+        return sorted(self.parent_refs, key=lambda pr: pr.get_inheritance_priority(), reverse=reverse)
 
     def _build_coded_prefix_tree(self):
         """Constructs the coded prefix tree of the services.
@@ -346,8 +353,7 @@ class DiagLayer:
         (a) SIDs for different services are the same like for service 1 and 2 (thus each leaf node is a list) and
         (b) one SID is the prefix of another SID like for service 3 and 4 (thus the constant `-1` key).
         """
-        services = list(filter(lambda s: isinstance(
-            s, DiagService), self._services))
+        services = [ s for s in self._services if isinstance(s, DiagService) ]
         prefix_tree = {}
         for s in services:
             # Compute prefixes for the request and all responses
@@ -437,84 +443,98 @@ class DiagLayer:
                 f"None of the services {possible_services} could parse {response.hex()}.")
         return decoded_messages
 
-    def get_receive_id(self):
+    def get_communication_parameter(self, cp_id: str) \
+        -> Optional[CommunicationParameterRef]:
+
+        cps = [ cp for cp in self.communication_parameters if cp.id_ref.ref_id == cp_id ]
+        if len(cps) > 1:
+            warnings.warn(f"Communication parameter `{cp_id}` specified more "
+                          f"than once. Using first occurence.", OdxWarning)
+        elif len(cps) == 0:
+            return None
+
+        return cps[0]
+
+    def get_receive_id(self) -> Optional[int]:
         """CAN ID to which the ECU listens for diagnostic messages"""
-        # TODO (?): When using the dict notation to access comparams,
-        #           wouldn't it be prettier to allow the dot in the id?
-        com_param = self.communication_parameters["ISO_15765_2__CP_UniqueRespIdTable"]
+        com_param = self.get_communication_parameter("ISO_15765_2.CP_UniqueRespIdTable")
+
         if com_param is None:
             return None
+
+        if self._enable_candela_workarounds:
+            # assume the parameter order used by CANdela studio.
+            # note that the parameter ordering actually used
+            # differs from the one of the COMPARAM fragment
+            # delivered by CANdela generated PDX files and that
+            # both are different from the one of the COMPARAM
+            # fragment included in the MCD2-D standard
+            try:
+                return int(com_param.value[2])
+            except ValueError:
+                return None
         else:
-            if self._enable_candela_workarounds:
-                # assume the parameter order used by CANdela studio.
-                # note that the parameter ordering actually used
-                # differs from the one of the COMPARAM fragment
-                # delivered by CANdela generated PDX files and that
-                # both are different from the one of the COMPARAM
-                # fragment included in the MCD2-D standard
-                try:
-                    return int(com_param.value[2])
-                except ValueError:
-                    return None
+            # assume the parameter order specified by the COMPARAM
+            # fragment of the ASAM MCD2-D standard.
+            return int(com_param.value[1])
 
-            else:
-                # assume the parameter order specified by the COMPARAM
-                # fragment of the ASAM MCD2-D standard.
-                return int(com_param.value[1])
-
-    def get_send_id(self):
+    def get_send_id(self) -> Optional[int]:
         """CAN ID to which the ECU sends replies to diagnostic messages"""
-        com_param = self.communication_parameters["ISO_15765_2__CP_UniqueRespIdTable"]
+        com_param = self.get_communication_parameter("ISO_15765_2.CP_UniqueRespIdTable")
+
         if com_param is None:
             return None
-        else:
-            if self._enable_candela_workarounds:
-                # assume the parameter order used by CANdela studio.
-                # note that the parameter odering actually used
-                # differs from the one of the COMPARAM fragment
-                # delivered by CANdela generated PDX files and that
-                # both are different from the one of the COMPARAM
-                # fragment included in the MCD2-D standard
-                return int(com_param.value[5])
-            else:
-                # assume the parameter order specified by the COMPARAM
-                # fragment of the ASAM MCD2-D standard.
-                return int(com_param.value[4])
 
-    def get_can_func_req_id(self):
+        if self._enable_candela_workarounds:
+            # assume the parameter order used by CANdela studio.
+            # note that the parameter odering actually used
+            # differs from the one of the COMPARAM fragment
+            # delivered by CANdela generated PDX files and that
+            # both are different from the one of the COMPARAM
+            # fragment included in the MCD2-D standard
+            return int(com_param.value[5])
+        else:
+            # assume the parameter order specified by the COMPARAM
+            # fragment of the ASAM MCD2-D standard.
+            return int(com_param.value[4])
+
+    def get_can_func_req_id(self) -> Optional[int]:
         """CAN Functional Request Id."""
-        com_param = self.communication_parameters["ISO_15765_2__CP_CanFuncReqId"]
+        com_param = self.get_communication_parameter("ISO_15765_2.CP_CanFuncReqId")
+
         if com_param is None:
             return None
-        else:
-            return int(com_param.value)
 
-    def get_logical_doip_address(self):
+        return int(com_param.value)
+
+    def get_logical_doip_address(self) -> Optional[int]:
         """The logical DoIP address of the ECU."""
-        com_param = self.communication_parameters["ISO_13400_2_DIS_2015__CP_UniqueRespIdTable"]
-        return int(com_param.value[0]) if com_param is not None else 0
+        com_param = self.get_communication_parameter("ISO_13400_2_DIS_2015.CP_UniqueRespIdTable")
 
-    def get_tester_present_time(self):
+        if com_param is None:
+            return None
+
+        return int(com_param.value[0])
+
+    def get_tester_present_time(self) -> Optional[float]:
         """Timeout on inactivity in seconds.
 
-        This is defined by the communication parameter "ISO_14230_3.CP_TesterPresentTime".
+        This is defined by the communication parameter "ISO_15765_3.CP_TesterPresentTime".
         If the variant does not define this parameter, the default value 3.0 is returned.
 
         Description of the comparam: "Time between a response and the next subsequent tester present message
         (if no other request is sent to this ECU) in case of physically addressed requests."
         """
-        cps = list(filter(lambda x: x.id_ref == "ISO_14230_3.CP_TesterPresentTime",
-                          self.communication_parameters))
-        if len(cps):
-            assert len(cps) == 1
-            idle_timeout = int(cps[0].value) / 1e6
-        else:
-            idle_timeout = 3.0  # default specified by the standard
-        return idle_timeout
+        com_param = self.get_communication_parameter("ISO_15765_3.CP_TesterPresentTime")
+
+        if com_param is None:
+            return 3.0  # default specified by the standard
+
+        return int(com_param.value) / 1e6
 
     def __repr__(self) -> str:
         return f"""DiagLayer(variant_type={self.variant_type},
-          id={repr(self.id)},
+          odx_id={repr(self.odx_id)},
           short_name={repr(self.short_name)},
           long_name={repr(self.long_name)},
           description={repr(self.description)},
@@ -532,8 +552,10 @@ class DiagLayer:
         return f"DiagLayer('{self.short_name}', type='{self.variant_type}')"
 
 
-def read_parent_ref_from_odx(et_element):
-    id_ref = make_ref(et_element)
+def read_parent_ref_from_odx(et_element, doc_frags: List[OdxDocFragment]) \
+    -> DiagLayer.ParentRef:
+    parent_ref = OdxLinkRef.from_et(et_element, doc_frags)
+    assert parent_ref is not None
 
     not_inherited_diag_comms = [el.get("SHORT-NAME")
                                 for el in et_element.iterfind("NOT-INHERITED-DIAG-COMMS/NOT-INHERITED-DIAG-COMM/DIAG-COMM-SNREF")]
@@ -542,77 +564,104 @@ def read_parent_ref_from_odx(et_element):
     ref_type = et_element.get(f"{xsi}type")
 
     return DiagLayer.ParentRef(
-        id_ref,
+        parent_ref,
         ref_type=ref_type,
         not_inherited_diag_comms=not_inherited_diag_comms,
         not_inherited_dops=not_inherited_dops
     )
 
 
-def read_diag_layer_from_odx(et_element, enable_candela_workarounds=True):
+def read_diag_layer_from_odx(et_element,
+                             doc_frags: List[OdxDocFragment],
+                             enable_candela_workarounds=True) \
+    -> DiagLayer:
 
-    # logger.info(et_element)
     variant_type = et_element.tag
 
-    id = et_element.get("ID")
     short_name = et_element.find("SHORT-NAME").text
-
-    long_name = et_element.find(
-        "LONG-NAME").text if et_element.find("LONG-NAME") is not None else None
+    long_name = et_element.findtext("LONG-NAME")
     description = read_description_from_odx(et_element.find("DESC"))
 
-    logger.info(f"Parsing {variant_type} '{short_name}' ...")
+    logger.info(f"Parsing diagnostic layer '{short_name}' "
+                f"of type {variant_type} ...")
+
+    # extend the applicable ODX "document fragments" for the diag layer objects
+    doc_frags = copy(doc_frags)
+    doc_frags.append(OdxDocFragment(short_name, "LAYER"))
+
+    odx_id = OdxLinkId.from_et(et_element, doc_frags)
 
     # Parse DiagServices
-    services = [read_diag_service_from_odx(service)
+    services = [read_diag_service_from_odx(service, doc_frags)
                 for service in et_element.iterfind("DIAG-COMMS/DIAG-SERVICE")]
-    diag_comm_refs = [make_ref(service)
-                      for service in et_element.iterfind("DIAG-COMMS/DIAG-COMM-REF")]
-    single_ecu_jobs = [read_single_ecu_job_from_odx(sej)
+    diag_comm_refs = []
+    for service in et_element.iterfind("DIAG-COMMS/DIAG-COMM-REF"):
+        ref = OdxLinkRef.from_et(service, doc_frags)
+        assert ref is not None
+        diag_comm_refs.append(ref)
+
+    single_ecu_jobs = [read_single_ecu_job_from_odx(sej, doc_frags)
                        for sej in et_element.iterfind("DIAG-COMMS/SINGLE-ECU-JOB")]
 
     # Parse ParentRefs
-    parent_refs = [read_parent_ref_from_odx(pr_el)
+    parent_refs = [read_parent_ref_from_odx(pr_el, doc_frags)
                    for pr_el in et_element.iterfind("PARENT-REFS/PARENT-REF")]
 
     # Parse communication parameter refs
-    com_params = [read_communication_param_ref_from_odx(el)
+    com_params = [read_communication_param_ref_from_odx(el, doc_frags)
                   for el in et_element.iterfind("COMPARAM-REFS/COMPARAM-REF")]
 
     # Parse Requests and Responses
-    requests = [read_structure_from_odx(rq)
-                for rq in et_element.iterfind("REQUESTS/REQUEST")]
-    positive_responses = [read_structure_from_odx(pr)
-                          for pr in et_element.iterfind("POS-RESPONSES/POS-RESPONSE")]
-    negative_responses = [read_structure_from_odx(nr)
-                          for nr in et_element.iterfind("NEG-RESPONSES/NEG-RESPONSE")]
+    requests = []
+    for rq_elem in et_element.iterfind("REQUESTS/REQUEST"):
+        rq = read_structure_from_odx(rq_elem, doc_frags)
+        assert isinstance(rq, Request)
+        requests.append(rq)
 
-    additional_audiences = [read_additional_audience_from_odx(el)
+    positive_responses = []
+    for pr_elem in et_element.iterfind("POS-RESPONSES/POS-RESPONSE"):
+        pr = read_structure_from_odx(pr_elem, doc_frags)
+        assert isinstance(pr, Response)
+        positive_responses.append(pr)
+
+    negative_responses = []
+    for nr_elem in et_element.iterfind("NEG-RESPONSES/NEG-RESPONSE"):
+        nr = read_structure_from_odx(nr_elem, doc_frags)
+        assert isinstance(nr, Response)
+        negative_responses.append(nr)
+
+    additional_audiences = [read_additional_audience_from_odx(el, doc_frags)
                             for el in et_element.iterfind("ADDITIONAL-AUDIENCES/ADDITIONAL-AUDIENCE")]
 
     functional_classes = [
-        read_functional_class_from_odx(el) for el in et_element.iterfind("FUNCT-CLASSS/FUNCT-CLASS")]
+        read_functional_class_from_odx(el, doc_frags)
+        for el in et_element.iterfind("FUNCT-CLASSS/FUNCT-CLASS")]
 
     states = [
-        read_state_from_odx(el) for el in et_element.iterfind("STATE-CHARTS/STATE-CHART/STATES/STATE")]
+        read_state_from_odx(el, doc_frags)
+        for el in et_element.iterfind("STATE-CHARTS/STATE-CHART/STATES/STATE")]
 
     state_transitions = [
-        read_state_transition_from_odx(el) for el in
-        et_element.iterfind("STATE-CHARTS/STATE-CHART/STATE-TRANSITIONS/STATE-TRANSITION")]
+        read_state_transition_from_odx(el, doc_frags)
+        for el in et_element.iterfind("STATE-CHARTS/"
+                                      "STATE-CHART/"
+                                      "STATE-TRANSITIONS/"
+                                      "STATE-TRANSITION")]
 
     if et_element.find("DIAG-DATA-DICTIONARY-SPEC"):
         diag_data_dictionary_spec = read_diag_data_dictionary_spec_from_odx(
-            et_element.find("DIAG-DATA-DICTIONARY-SPEC"))
+            et_element.find("DIAG-DATA-DICTIONARY-SPEC"), doc_frags)
     else:
         diag_data_dictionary_spec = None
 
-    import_refs = [make_ref(ref) for ref in et_element.iterfind("IMPORT-REFS/IMPORT-REF")]
+    import_refs = [OdxLinkRef.from_et(ref, doc_frags)
+                   for ref in et_element.iterfind("IMPORT-REFS/IMPORT-REF")]
 
     # TODO: Are UNIT-SPEC and SDGS needed?
 
     # Create DiagLayer
     dl = DiagLayer(variant_type,
-                   id,
+                   odx_id,
                    short_name,
                    long_name=long_name,
                    description=description,
@@ -638,7 +687,7 @@ def read_diag_layer_from_odx(et_element, enable_candela_workarounds=True):
 
 class DiagLayerContainer:
     def __init__(self,
-                 id,
+                 odx_id,
                  short_name,
                  long_name=None,
                  description=None,
@@ -650,7 +699,7 @@ class DiagLayerContainer:
                  base_variants=[],
                  ecu_variants=[]
                  ):
-        self.id = id
+        self.odx_id = odx_id
         self.short_name = short_name
         self.long_name = long_name
         self.description = description
@@ -664,27 +713,29 @@ class DiagLayerContainer:
         self.ecu_variants = ecu_variants
 
         self._diag_layers = NamedItemList[DiagLayer](
-            lambda dop: dop.short_name,
+            short_name_as_id,
             list(chain(self.ecu_shared_datas, self.protocols, self.functional_groups, self.base_variants, self.ecu_variants)))
 
-    def _build_id_lookup(self):
+    def _build_odxlinks(self):
         result = {}
+        result[self.odx_id] = self
+
         if self.admin_data is not None:
-            result.update(self.admin_data._build_id_lookup())
+            result.update(self.admin_data._build_odxlinks())
 
         if self.company_datas is not None:
             for cd in self.company_datas:
-                result.update(cd._build_id_lookup())
+                result.update(cd._build_odxlinks())
 
         return result
 
-    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
+    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
         if self.admin_data is not None:
-            self.admin_data._resolve_references(id_lookup)
+            self.admin_data._resolve_references(odxlinks)
 
         if self.company_datas is not None:
             for cd in self.company_datas:
-                cd._resolve_references(id_lookup)
+                cd._resolve_references(odxlinks)
 
 
     @property
@@ -702,27 +753,41 @@ class DiagLayerContainer:
 
 
 def read_diag_layer_container_from_odx(et_element, enable_candela_workarounds=True):
-    id = et_element.get("ID")
-    short_name = et_element.find("SHORT-NAME").text
-    try:
-        long_name = et_element.find("LONG-NAME").text
-    except:
-        long_name = None
+    short_name = et_element.findtext("SHORT-NAME")
+    assert short_name is not None
+    long_name = et_element.findtext("LONG-NAME")
+
+    # create the current ODX "document fragment" (description of the
+    # current document for references and IDs)
+    doc_frags = [ OdxDocFragment(short_name, "CONTAINER") ]
+
+    odx_id = OdxLinkId.from_et(et_element, doc_frags)
+    assert odx_id is not None
     description = read_description_from_odx(et_element.find("DESC"))
-    admin_data = read_admin_data_from_odx(et_element.find("ADMIN-DATA"))
-    company_datas = read_company_datas_from_odx(et_element.find("COMPANY-DATAS"))
-    ecu_shared_datas = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
+    admin_data = read_admin_data_from_odx(et_element.find("ADMIN-DATA"), doc_frags)
+    company_datas = read_company_datas_from_odx(et_element.find("COMPANY-DATAS"), doc_frags)
+    ecu_shared_datas = [read_diag_layer_from_odx(dl_element,
+                                                 doc_frags,
+                                                 enable_candela_workarounds=enable_candela_workarounds)
                         for dl_element in et_element.iterfind("ECU-SHARED-DATAS/ECU-SHARED-DATA")]
-    protocols = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
+    protocols = [read_diag_layer_from_odx(dl_element,
+                                          doc_frags,
+                                          enable_candela_workarounds=enable_candela_workarounds)
                  for dl_element in et_element.iterfind("PROTOCOLS/PROTOCOL")]
-    functional_groups = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
+    functional_groups = [read_diag_layer_from_odx(dl_element,
+                                                  doc_frags,
+                                                  enable_candela_workarounds=enable_candela_workarounds)
                          for dl_element in et_element.iterfind("FUNCTIONAL-GROUPS/FUNCTIONAL-GROUP")]
-    base_variants = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
+    base_variants = [read_diag_layer_from_odx(dl_element,
+                                              doc_frags,
+                                              enable_candela_workarounds=enable_candela_workarounds)
                      for dl_element in et_element.iterfind("BASE-VARIANTS/BASE-VARIANT")]
-    ecu_variants = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
+    ecu_variants = [read_diag_layer_from_odx(dl_element,
+                                             doc_frags,
+                                             enable_candela_workarounds=enable_candela_workarounds)
                     for dl_element in et_element.iterfind("ECU-VARIANTS/ECU-VARIANT")]
 
-    return DiagLayerContainer(id,
+    return DiagLayerContainer(odx_id,
                               short_name,
                               long_name=long_name,
                               description=description,
