@@ -1,19 +1,24 @@
 # SPDX-License-Identifier: MIT
 # Copyright (c) 2022 MBition GmbH
 
-from typing import List, Union
-from .utils import make_ref, read_description_from_odx
+from typing import TYPE_CHECKING, Any, List, Union
+
+from .utils import read_description_from_odx
+from .odxlink import OdxLinkRef, OdxLinkId, OdxDocFragment, OdxLinkDatabase
 from .structures import BasicStructure
 from .dataobjectproperty import DopBase
 from .decodestate import DecodeState
 from .encodestate import EncodeState
 
+if TYPE_CHECKING:
+    from .diaglayer import DiagLayer
+
 
 class EndOfPduField(DopBase):
     """ End of PDU fields are structures that are repeated until the end of the PDU """
 
     def __init__(self,
-                 id,
+                 odx_id,
                  short_name,
                  structure=None,
                  structure_ref=None,
@@ -23,7 +28,7 @@ class EndOfPduField(DopBase):
                  is_visible=False,
                  long_name=None,
                  description=None):
-        super().__init__(id, short_name, long_name=long_name,
+        super().__init__(odx_id, short_name, long_name=long_name,
                          description=description, is_visible=is_visible)
 
         self.structure_snref = structure_snref
@@ -31,7 +36,7 @@ class EndOfPduField(DopBase):
 
         self._structure = structure
         if structure:
-            self.structure_ref = structure.id
+            self.structure_ref = OdxLinkRef.from_id(structure.odx_id)
         assert self.structure_ref or self.structure_snref
 
         self.min_number_of_items = min_number_of_items
@@ -51,7 +56,7 @@ class EndOfPduField(DopBase):
     def convert_physical_to_internal(self, physical_value):
         return self.structure.convert_physical_to_internal(physical_value)
 
-    def convert_physical_to_bytes(self, physical_value: Union[dict, List[dict]], encode_state: EncodeState, bit_position=0):
+    def convert_physical_to_bytes(self, physical_value: Union[dict, List[dict]], encode_state: EncodeState, bit_position: int = 0):
         assert bit_position == 0, "End of PDU field must be byte aligned. Is there an error in reading the .odx?"
         if isinstance(physical_value, dict):
             # If the value is given as a dict, the End of PDU field behaves like the underlying structure.
@@ -65,7 +70,7 @@ class EndOfPduField(DopBase):
                 coded_rpc += self.structure.convert_physical_to_bytes(value, encode_state)
             return coded_rpc
 
-    def convert_bytes_to_physical(self, decode_state: DecodeState, bit_position=0):
+    def convert_bytes_to_physical(self, decode_state: DecodeState, bit_position: int = 0):
         next_byte_position = decode_state.next_byte_position
         byte_code = decode_state.coded_message
 
@@ -83,33 +88,38 @@ class EndOfPduField(DopBase):
 
         return value, next_byte_position
 
-    def _resolve_references(self, parent_dl, id_lookup):
+    def _resolve_references(self,
+                            parent_dl: "DiagLayer",
+                            odxlinks: OdxLinkDatabase) \
+     -> None:
         """Recursively resolve any references (odxlinks or sn-refs)
         """
         if self.structure_ref is not None:
-            self._structure = id_lookup[self.structure_ref]
+            self._structure = odxlinks.resolve(self.structure_ref)
         else:
             self._structure = parent_dl.data_object_properties[self.structure_snref]
 
     def __repr__(self) -> str:
-        return f"EndOfPduField(short_name='{self.short_name}', ref='{self.structure.id}')"
+        return f"EndOfPduField(short_name='{self.short_name}', ref='{self.structure.odx_id}')"
 
     def __str__(self):
         return "\n".join([
-            f"EndOfPduField(short_name='{self.short_name}', ref='{self.structure.id}')"
+            f"EndOfPduField(short_name='{self.short_name}', ref='{self.structure.odx_id}')"
         ] + [
             " " + str(self.structure).replace("\n", "\n ")
         ])
 
 
-def read_end_of_pdu_field_from_odx(et_element):
-    id = et_element.get("ID")
+def read_end_of_pdu_field_from_odx(et_element, doc_frags: List[OdxDocFragment]) \
+    -> EndOfPduField:
+    odx_id = OdxLinkId.from_et(et_element, doc_frags)
+    assert odx_id is not None
     short_name = et_element.find("SHORT-NAME").text
-    long_name = et_element.find(
-        "LONG-NAME").text if et_element.find("LONG-NAME") is not None else None
+    long_name = et_element.findtext("LONG-NAME")
     description = read_description_from_odx(et_element.find("DESC"))
 
-    structure_ref = make_ref(et_element.find("BASIC-STRUCTURE-REF"))
+    structure_ref = OdxLinkRef.from_et(et_element.find("BASIC-STRUCTURE-REF"), doc_frags)
+    assert structure_ref is not None
     structure_snref = None
 
     if et_element.find("BASIC-STRUCTURE-SNREF") is not None:
@@ -118,7 +128,7 @@ def read_end_of_pdu_field_from_odx(et_element):
             "BASIC-STRUCTURE-SNREF").get("SHORT-NAME")
 
     if et_element.find("ENV-DATA-DESC-REF") is not None:
-        structure_ref = make_ref(et_element.get("ENV-DATA-DESC-REF"))
+        structure_ref = OdxLinkRef.from_et(et_element.get("ENV-DATA-DESC-REF"), doc_frags)
         structure_snref = None
 
     if et_element.find("ENV-DATA-DESC-SNREF") is not None:
@@ -142,7 +152,7 @@ def read_end_of_pdu_field_from_odx(et_element):
         is_visible = is_visible == "true"
     else:
         is_visible = False
-    eopf = EndOfPduField(id,
+    eopf = EndOfPduField(odx_id,
                          short_name,
                          long_name=long_name,
                          description=description,
