@@ -5,7 +5,8 @@ import abc
 from dataclasses import dataclass
 from typing import Optional, List, Dict, Any, Iterable
 
-from odxtools.utils import read_description_from_odx, make_ref
+from .utils import read_description_from_odx
+from .odxlink import OdxLinkRef, OdxLinkId, OdxLinkDatabase, OdxDocFragment
 
 from .dataobjectproperty import DopBase
 from .globals import logger
@@ -14,8 +15,8 @@ from .globals import logger
 class TableBase(abc.ABC):
     """ Base class for all Tables."""
 
-    def __init__(self, id: str, short_name: str, long_name=None):
-        self.id = id
+    def __init__(self, odx_id: OdxLinkId, short_name: str, long_name=None):
+        self.odx_id = odx_id
         self.short_name = short_name
         self.long_name = long_name
 
@@ -24,12 +25,12 @@ class TableBase(abc.ABC):
 class TableRow:
     """This class represents a TABLE-ROW."""
 
-    id: str
+    odx_id: OdxLinkId
     short_name: str
     long_name: str
     key: int
-    structure_ref: Optional[str] = None
-    dop_ref: Optional[str] = None
+    structure_ref: Optional[OdxLinkRef] = None
+    dop_ref: Optional[OdxLinkRef] = None
     description: Optional[str] = None
     semantic: Optional[str] = None
 
@@ -37,19 +38,11 @@ class TableRow:
         self._structure: Optional[DopBase] = None
         self._dop: Optional[DopBase] = None
 
-    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
+    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
         if self.structure_ref is not None:
-            self._structure = id_lookup.get(self.structure_ref)
-            if self._structure is None:
-                logger.warning(
-                    f"STRUCTURE-REF '{self.structure_ref}' could not be resolved in TABLE-ROW."
-                )
+            self._structure = odxlinks.resolve(self.structure_ref)
         if self.dop_ref is not None:
-            self._dop = id_lookup.get(self.dop_ref)
-            if self._dop is None:
-                logger.warning(
-                    f"DATA-OBJECT-PROP-REF '{self.dop_ref}' could not be resolved in TABLE-ROW."
-                )
+            self._dop = odxlinks.resolve(self.dop_ref)
 
     @property
     def structure(self) -> Optional[DopBase]:
@@ -79,17 +72,17 @@ class Table(TableBase):
 
     def __init__(
         self,
-        id: str,
+        odx_id: OdxLinkId,
         short_name: str,
         table_rows: List[TableRow],
-        table_row_refs: Optional[List[str]] = None,
+        table_row_refs: Optional[List[OdxLinkRef]] = None,
         long_name: Optional[str] = None,
-        key_dop_ref: Optional[str] = None,
+        key_dop_ref: Optional[OdxLinkRef] = None,
         description: Optional[str] = None,
         semantic: Optional[str] = None,
     ):
         super().__init__(
-            id=id, short_name=short_name, long_name=long_name
+            odx_id=odx_id, short_name=short_name, long_name=long_name
         )
         self._local_table_rows = table_rows
         self._ref_table_rows: List[TableRow] = []
@@ -109,24 +102,21 @@ class Table(TableBase):
         """The table rows (both local and referenced) in this table."""
         return self._local_table_rows + self._ref_table_rows
 
-    def _build_id_lookup(self):
-        id_lookup = {}
-        id_lookup.update({table_row.id: table_row for table_row in self.table_rows})
-        return id_lookup
+    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
+        odxlinks = {}
+        odxlinks.update({table_row.odx_id: table_row for table_row in self.table_rows})
+        return odxlinks
 
-    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
+    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
         if self.key_dop_ref is not None:
-            self._key_dop = id_lookup.get(self.key_dop_ref)
-            if self._key_dop is None:
-                logger.warning(
-                    f"KEY-DOP-REF '{self.key_dop_ref!r}' could not be resolved in TABLE."
-                )
+            self._key_dop = odxlinks.resolve(self.key_dop_ref)
+
         for table_row in self._local_table_rows:
-            table_row._resolve_references(id_lookup)
+            table_row._resolve_references(odxlinks)
 
         self._ref_table_rows = []
         for ref in self._table_row_refs:
-            tr = id_lookup.get(ref)
+            tr = odxlinks.resolve(ref)
             if isinstance(tr, TableRow):
                 self._ref_table_rows.append(tr)
 
@@ -140,10 +130,10 @@ class Table(TableBase):
         )
 
 
-def _get_common_props(et_element):
+def _get_common_props(et_element, doc_frags: List[OdxDocFragment]):
     description = read_description_from_odx(et_element.find("DESC"))
     return dict(
-        id=et_element.get("ID"),
+        odx_id=OdxLinkId.from_et(et_element, doc_frags),
         short_name=et_element.findtext("SHORT-NAME"),
         long_name=et_element.findtext("LONG-NAME"),
         semantic=et_element.get("SEMANTIC"),
@@ -151,47 +141,47 @@ def _get_common_props(et_element):
     )
 
 
-def read_table_row_from_odx(et_element):
+def read_table_row_from_odx(et_element, doc_frags: List[OdxDocFragment]) \
+    -> TableRow:
     """Reads a TABLE-ROW."""
     key = et_element.find("KEY").text
     structure_ref = None
     if et_element.find("STRUCTURE-REF") is not None:
-        structure_ref = make_ref(et_element.find("STRUCTURE-REF"))
+        structure_ref = OdxLinkRef.from_et(et_element.find("STRUCTURE-REF"), doc_frags)
     dop_ref = None
     if et_element.find("DATA-OBJECT-PROP-REF") is not None:
-        dop_ref = make_ref(et_element.find("DATA-OBJECT-PROP-REF"))
+        dop_ref = OdxLinkRef.from_et(et_element.find("DATA-OBJECT-PROP-REF"), doc_frags)
 
-    table_row = TableRow(
+    return TableRow(
         key=key,
         structure_ref=structure_ref,
         dop_ref=dop_ref,
-        **_get_common_props(et_element)
+        **_get_common_props(et_element, doc_frags)
     )
 
-    return table_row
 
-
-def read_table_from_odx(et_element):
+def read_table_from_odx(et_element, doc_frags: List[OdxDocFragment]) \
+    -> Table:
     """Reads a TABLE."""
     short_name = et_element.find("SHORT-NAME").text
     key_dop_ref = None
     if et_element.find("KEY-DOP-REF") is not None:
-        key_dop_ref = make_ref(et_element.find("KEY-DOP-REF"))
+        key_dop_ref = OdxLinkRef.from_et(et_element.find("KEY-DOP-REF"), doc_frags)
     logger.debug("Parsing TABLE " + short_name)
 
     table_rows = [
-        read_table_row_from_odx(el) for el in et_element.iterfind("TABLE-ROW")
+        read_table_row_from_odx(el, doc_frags) for el in et_element.iterfind("TABLE-ROW")
     ]
 
-    table_row_refs = [
-        make_ref(el) for el in et_element.iterfind("TABLE-ROW-REF")
-    ]
+    table_row_refs = [ ]
+    for el in et_element.iterfind("TABLE-ROW-REF"):
+        ref = OdxLinkRef.from_et(el, doc_frags)
+        assert ref is not None
+        table_row_refs.append(ref)
 
-    table = Table(
+    return Table(
         key_dop_ref=key_dop_ref,
         table_rows=table_rows,
         table_row_refs=table_row_refs,
-        **_get_common_props(et_element)
+        **_get_common_props(et_element, doc_frags)
     )
-
-    return table
